import shutil
import logging
import tempfile
import threading
from normality import safe_filename

from servicelayer.archive.archive import Archive
from servicelayer.archive.util import ensure_path

log = logging.getLogger(__name__)


class VirtualArchive(Archive):

    def __init__(self, base_name):
        self.local = threading.local()
        self.base_name = base_name

    def _get_local_prefix(self, content_hash, temp_path=None):
        """Determine a temporary path for the file on the local file
        system."""
        if temp_path is None:
            if not hasattr(self.local, 'dir'):
                self.local.dir = tempfile.mkdtemp(prefix=self.base_name)
            temp_path = self.local.dir
        return ensure_path(temp_path).joinpath('%s.sl' % content_hash)

    def _local_path(self, content_hash, file_name, temp_path):
        path = self._get_local_prefix(content_hash, temp_path=temp_path)
        path.mkdir(parents=True, exist_ok=True)
        return path.joinpath(safe_filename(file_name, default='data'))

    def cleanup_file(self, content_hash, temp_path=None):
        """Delete the local cached version of the file."""
        if content_hash is None:
            return
        path = self._get_local_prefix(content_hash, temp_path=temp_path)
        try:
            shutil.rmtree(path, ignore_errors=True)
        except Exception:
            pass



from unittest import TestCase

from servicelayer.cache import get_fakeredis
from servicelayer.jobs import Job, Stage, Task, Dataset


class ProcessTest(TestCase):

    def setUp(self):
        self.conn = get_fakeredis()
        self.conn.flushall()
        self.dataset = 'test_1'

    def test_job_queue(self):
        job = Job.create(self.conn, self.dataset)
        stage = job.get_stage('ingest')
        status = stage.get_status()
        assert status['pending'] == 0
        assert status['finished'] == 0
        assert job.is_done()
        stage.queue({'test': 'foo'}, {})
        status = job.get_status()
        assert status['pending'] == 1
        assert status['finished'] == 0
        assert status['running'] == 0
        assert not job.is_done()
        task = Stage.get_task(self.conn, 'ingest',
                              timeout=None)
        assert task.job.dataset.name == job.dataset.name
        assert task.payload['test'] == 'foo'
        status = job.get_status()
        assert status['pending'] == 0
        assert status['running'] == 1
        assert status['finished'] == 0
        assert not job.is_done()
        task.done()
        status = job.get_status()
        assert status['pending'] == 0
        assert status['running'] == 0
        assert status['finished'] == 1
        assert job.is_done()

    def test_queue_clear(self):
        job = Job.create(self.conn, self.dataset)
        stage = job.get_stage('ingest')

        stage.queue({'test': 'foo'}, {})
        status = stage.get_status()
        assert status['pending'] == 1
        job.dataset.cancel()
        status = stage.get_status()
        assert status['pending'] == 0

        stage.queue({'test': 'foo'}, {})
        status = stage.get_status()
        assert status['pending'] == 1
        job.remove()
        status = stage.get_status()
        assert status['pending'] == 0

    def test_fake_finished(self):
        job = Job.create(self.conn, self.dataset)
        stage = job.get_stage('ingest')
        status = stage.get_status()
        assert status['finished'] == 0
        stage.report_finished(500)
        status = stage.get_status()
        assert status['finished'] == 500
        status = job.dataset.get_status()
        assert status['finished'] == 500, status

    def test_fetch_multiple_task(self):
        job = Job.create(self.conn, self.dataset)
        stage = job.get_stage('ingest')
        stage.queue({'test': 'foo'}, {})
        stage.queue({'test': 'bar'}, {})
        status = job.get_status()
        assert status['pending'] == 2
        tasks = list(stage.get_tasks(limit=5))
        assert len(tasks) == 2
        for task in tasks:
            assert isinstance(task, Task)
        assert tasks[0].payload == {'test': 'foo'}
        assert tasks[1].payload == {'test': 'bar'}
        job.dataset.cancel()

    def test_active_dataset_status(self):
        job = Job.create(self.conn, self.dataset)
        stage = job.get_stage('ingest')
        stage.queue({'test': 'foo'}, {})
        stage.queue({'test': 'bar'}, {})
        status = Dataset.get_active_dataset_status(self.conn)
        assert len(status['datasets']) == 1
        assert status['total'] == 1
        assert status['datasets']['test_1']['pending'] == 2
        job.dataset.cancel()
        status = Dataset.get_active_dataset_status(self.conn)
        assert status['datasets'] == {}
        assert status['total'] == 0



import os
import abc


class Archive(object):
    __metaclass__ = abc.ABCMeta

    def _get_prefix(self, content_hash):
        if content_hash is not None:
            return os.path.join(content_hash[:2],
                                content_hash[2:4],
                                content_hash[4:6],
                                content_hash)

    def upgrade(self):
        """Run maintenance on the store."""
        pass

    @abc.abstractmethod
    def archive_file(self, file_path, content_hash=None, mime_type=None):
        """Import the given file into the archive."""
        pass

    @abc.abstractmethod
    def load_file(self, content_hash, file_name=None, temp_path=None):
        pass

    def cleanup_file(self, content_hash, temp_path=None):
        pass

    def generate_url(self, content_hash, file_name=None, mime_type=None):
        return None



import os
import logging
from datetime import datetime, timedelta
from google.cloud.storage import Blob
from google.cloud.storage.client import Client
from google.api_core.exceptions import TooManyRequests, InternalServerError
from google.api_core.exceptions import ServiceUnavailable
from google.resumable_media.common import DataCorruption, InvalidResponse
from normality import safe_filename

from servicelayer.archive.virtual import VirtualArchive
from servicelayer.archive.util import checksum, ensure_path, ensure_posix_path
from servicelayer.util import service_retries, backoff

log = logging.getLogger(__name__)
FAILURES = (TooManyRequests, InternalServerError, ServiceUnavailable,
            DataCorruption, InvalidResponse)


class GoogleStorageArchive(VirtualArchive):
    TIMEOUT = 84600

    def __init__(self, bucket=None):
        super(GoogleStorageArchive, self).__init__(bucket)
        self.client = Client()
        log.info("Archive: gs://%s", bucket)

        self.bucket = self.client.lookup_bucket(bucket)
        if self.bucket is None:
            self.bucket = self.client.create_bucket(bucket)
            self.upgrade()

    def upgrade(self):
        # 'Accept-Ranges',
        # 'Content-Encoding',
        # 'Content-Length',
        # 'Content-Range',
        # 'Cache-Control',
        # 'Content-Language',
        # 'Content-Type',
        # 'Expires',
        # 'Last-Modified',
        # 'Pragma',
        # 'Range',
        # 'Date',
        policy = {
            "origin": ['*'],
            "method": ['GET', 'HEAD', 'OPTIONS'],
            "responseHeader": ['*'],
            "maxAgeSeconds": self.TIMEOUT
        }
        self.bucket.cors = [policy]
        self.bucket.update()

    def _locate_contenthash(self, content_hash):
        """Check if a file with the given hash exists on S3."""
        if content_hash is None:
            return
        prefix = self._get_prefix(content_hash)
        if prefix is None:
            return

        # First, check the standard file name:
        blob = Blob(os.path.join(prefix, 'data'), self.bucket)
        if blob.exists():
            return blob

        # Second, iterate over all file names:
        for blob in self.bucket.list_blobs(max_results=1, prefix=prefix):
            return blob

    def _locate_key(self, key):
        if key is None:
            return
        blob = Blob(key, self.bucket)
        if blob.exists():
            return blob

    def archive_file(self, file_path, content_hash=None, mime_type=None):
        """Store the file located at the given path on Google, based on a path
        made up from its SHA1 content hash."""
        file_path = ensure_path(file_path)
        if content_hash is None:
            content_hash = checksum(file_path)

        if content_hash is None:
            return

        file_path = ensure_posix_path(file_path)
        for attempt in service_retries():
            try:
                # blob = self._locate_contenthash(content_hash)
                # if blob is not None:
                #     return content_hash

                path = os.path.join(self._get_prefix(content_hash), 'data')
                blob = Blob(path, self.bucket)
                blob.upload_from_filename(file_path, content_type=mime_type)
                return content_hash
            except FAILURES:
                log.exception("Store error in GS")
                backoff(failures=attempt)

    def load_file(self, content_hash, file_name=None, temp_path=None):
        """Retrieve a file from Google storage and put it onto the local file
        system for further processing."""
        for attempt in service_retries():
            try:
                blob = self._locate_contenthash(content_hash)
                if blob is not None:
                    path = self._local_path(content_hash, file_name, temp_path)
                    blob.download_to_filename(path)
                    return path
            except FAILURES:
                log.exception("Load error in GS")
                backoff(failures=attempt)

        # Returns None for "persistent error" as well as "file not found" :/
        log.debug("[%s] not found, or the backend is down.", content_hash)

    def generate_url(self, content_hash, file_name=None, mime_type=None):
        blob = self._locate_contenthash(content_hash)
        if blob is None:
            return
        disposition = None
        if file_name is not None:
            disposition = 'inline; filename=%s' % file_name
        expire = datetime.utcnow() + timedelta(seconds=self.TIMEOUT)
        return blob.generate_signed_url(expire,
                                        response_type=mime_type,
                                        response_disposition=disposition)

    def publish(self, namespace, file_path, mime_type=None):
        file_path = ensure_posix_path(file_path)
        file_name = safe_filename(file_path, default='data')
        store_path = '{0}/{1}'.format(namespace, file_name)

        for attempt in service_retries():
            try:
                blob = Blob(store_path, self.bucket)
                blob.upload_from_filename(file_path, content_type=mime_type)
            except FAILURES:
                log.exception("Store error in GS")
                backoff(failures=attempt)

    def delete_publication(self, namespace, file_name):
        key = '{0}/{1}'.format(namespace, file_name)
        for attempt in service_retries():
            try:
                blob = self._locate_key(key)
                if blob is not None:
                    blob.delete()
            except FAILURES:
                log.exception("Load error in GS")
                backoff(failures=attempt)

        log.warn("[%s] not found, or the backend is down.", key)

    def generate_publication_url(self, namespace, file_name, mime_type=None,
                                 expire=None):
        key = '{0}/{1}'.format(namespace, file_name)
        blob = self._locate_key(key)
        if blob is None:
            return
        disposition = 'attachment; filename=%s' % file_name
        expire = expire or self.TIMEOUT
        expire = datetime.utcnow() + timedelta(seconds=expire)
        return blob.generate_signed_url(expire,
                                        response_type=mime_type,
                                        response_disposition=disposition)



from unittest import TestCase

from servicelayer.extensions import get_extensions


class ExtensionTest(TestCase):

    def test_extensions(self):
        exts = get_extensions('servicelayer.test')
        assert len(exts), exts
        assert get_extensions in exts, exts



from datetime import datetime
from normality import stringify

from servicelayer.jobs import Job
from servicelayer.settings import WORKER_REPORTING


OP_REPORT = 'report'


class Status:
    START = 'start'
    SUCCESS = 'success'
    ERROR = 'error'


class Reporter(object):
    """A reporter will create processing status updates."""

    def __init__(self, task=None, stage=None, operation=None):
        self.task = task
        self.stage = stage
        self.operation = operation

    def start(self, **data):
        self.handle(status=Status.START, **data)

    def end(self, **data):
        self.handle(status=Status.SUCCESS, **data)

    def error(self, exception, **data):
        self.handle(status=Status.ERROR, exception=exception, **data)

    def handle(self, status, operation=None, exception=None, task=None, **payload):
        """Report a processing event that may be related to a task."""
        if not WORKER_REPORTING:
            return

        task = task or self.task
        if task is not None:
            payload['task'] = task.serialize()
            stage = task.stage
        else:
            stage = self.stage
        dataset = stage.job.dataset.name
        job_id = stage.job.id
        operation = operation or stage.stage

        now = datetime.utcnow()
        payload.update({
            'dataset': dataset,
            'operation': operation,
            'job': job_id,
            'status': status,
            'updated_at': now,
            '%s_at' % status: now,
            'has_error': False,
        })

        if exception is not None:
            payload.update({
                'status': Status.ERROR,
                'has_error': True,
                'error_name': exception.__class__.__name__,
                'error_msg': stringify(exception)
            })

        job = Job(stage.conn, dataset, job_id)
        stage = job.get_stage(OP_REPORT)
        stage.queue(payload)



import shutil
import tempfile
from unittest import TestCase

from servicelayer.archive import init_archive
from servicelayer.archive.util import checksum, ensure_path


class FileArchiveTest(TestCase):

    def setUp(self):
        tempdir = ensure_path(tempfile.gettempdir())
        self.path = tempdir.joinpath('sltest')
        self.archive = init_archive('file', path=self.path)
        self.file = ensure_path(__file__)

    def tearDown(self):
        if self.path.exists():
            shutil.rmtree(self.path)

    def test_basic_archive(self):
        checksum_ = checksum(self.file)
        assert checksum_ is not None, checksum_
        out = self.archive.archive_file(self.file)
        assert checksum_ == out, (checksum_, out)
        out2 = self.archive.archive_file(self.file)
        assert out == out2, (out, out2)

    def test_basic_archive_with_checksum(self):
        checksum_ = 'banana'
        out = self.archive.archive_file(self.file, checksum_)
        assert checksum_ == out, (checksum_, out)

    def test_generate_url(self):
        out = self.archive.archive_file(self.file)
        url = self.archive.generate_url(out)
        assert url is None, url

    def test_load_file(self):
        out = self.archive.archive_file(self.file)
        path = self.archive.load_file(out)
        assert path is not None, path
        assert self.path in path.parents
        assert path.is_file(), path

    def test_cleanup_file(self):
        out = self.archive.archive_file(self.file)
        self.archive.cleanup_file(out)
        path = self.archive.load_file(out)
        assert path.is_file(), path

    def test_publication(self):
        self.archive.publish('banana', self.file, 'text/plain')
        path = self.archive.load_publication('banana', 'test_file.py')
        assert path is not None, path
        assert path.is_file(), path
        assert path.name == 'test_file.py', path.name
        assert path.parent.name == 'banana', path.parent
        self.archive.delete_publication('banana', 'test_file.py')
        assert not path.exists()



from hashlib import sha1
from pathlib import Path

BUF_SIZE = 1024 * 1024 * 16


def ensure_path(file_path):
    if file_path is None or isinstance(file_path, Path):
        return file_path
    return Path(file_path).resolve()


def ensure_posix_path(file_path):
    if isinstance(file_path, Path):
        file_path = file_path.as_posix()
    return file_path


def checksum(file_name):
    """Generate a hash for a given file name."""
    file_name = ensure_path(file_name)
    if file_name is not None and file_name.is_file():
        digest = sha1()
        with open(file_name, 'rb') as fh:
            while True:
                block = fh.read(BUF_SIZE)
                if not block:
                    break
                digest.update(block)
        return str(digest.hexdigest())



import boto3
import logging
from botocore.exceptions import ClientError
from normality import safe_filename

from servicelayer import settings
from servicelayer.archive.virtual import VirtualArchive
from servicelayer.archive.util import checksum, ensure_path

log = logging.getLogger(__name__)


class S3Archive(VirtualArchive):
    TIMEOUT = 84600

    def __init__(self, bucket=None):
        super(S3Archive, self).__init__(bucket)
        key_id = settings.AWS_KEY_ID
        secret_key = settings.AWS_SECRET_KEY
        self.client = boto3.client('s3',
                                   region_name=settings.AWS_REGION,
                                   aws_access_key_id=key_id,
                                   aws_secret_access_key=secret_key)
        # config=Config(signature_version='s3v4'))
        self.bucket = bucket
        log.info("Archive: s3://%s", bucket)

        try:
            self.client.head_bucket(Bucket=bucket)
        except ClientError as e:
            error_code = int(e.response['Error']['Code'])
            if error_code == 404:
                self.client.create_bucket(Bucket=bucket)
                self.upgrade()
            else:
                log.exception("Could not check bucket")

    def upgrade(self):
        # Make sure bucket policy is set correctly.
        config = {
            'CORSRules': [
                {
                    'AllowedMethods': ['GET'],
                    'AllowedOrigins': ['*'],
                    'AllowedHeaders': ['*'],
                    'ExposeHeaders': ['Accept-Ranges', 'Content-Encoding',
                                      'Content-Length', 'Content-Range'],
                    'MaxAgeSeconds': self.TIMEOUT
                }
            ]
        }
        try:
            self.client.put_bucket_cors(Bucket=self.bucket,
                                        CORSConfiguration=config)
        except ClientError:
            log.exception("Could not update CORS")

    def _locate_key(self, content_hash=None, prefix=None):
        """Check if a file with the given hash exists on S3."""
        if prefix is None:
            if content_hash is None:
                return
            prefix = self._get_prefix(content_hash)
            if prefix is None:
                return
        res = self.client.list_objects(MaxKeys=1,
                                       Bucket=self.bucket,
                                       Prefix=prefix)
        for obj in res.get('Contents', []):
            return obj.get('Key')

    def archive_file(self, file_path, content_hash=None, mime_type=None):
        """Store the file located at the given path on S3, based on a path
        made up from its SHA1 content hash."""
        file_path = ensure_path(file_path)
        if content_hash is None:
            content_hash = checksum(file_path)

        # if content_hash is None:
        #     return

        obj = self._locate_key(content_hash)
        if obj is not None:
            return content_hash

        path = '{}/data'.format(self._get_prefix(content_hash))
        extra = {}
        if mime_type is not None:
            extra['ContentType'] = mime_type
        with open(file_path, 'rb') as fh:
            self.client.upload_fileobj(fh, self.bucket, str(path),
                                       ExtraArgs=extra)
        return content_hash

    def load_file(self, content_hash, file_name=None, temp_path=None):
        """Retrieve a file from S3 storage and put it onto the local file
        system for further processing."""
        key = self._locate_key(content_hash)
        if key is not None:
            path = self._local_path(content_hash, file_name, temp_path)
            self.client.download_file(self.bucket, key, str(path))
            return path

    def generate_url(self, content_hash, file_name=None, mime_type=None):
        key = self._locate_key(content_hash)
        if key is None:
            return
        params = {
            'Bucket': self.bucket,
            'Key': key
        }
        if mime_type is not None:
            params['ResponseContentType'] = mime_type
        if file_name is not None:
            disposition = 'inline; filename=%s' % file_name
            params['ResponseContentDisposition'] = disposition
        return self.client.generate_presigned_url('get_object',
                                                  Params=params,
                                                  ExpiresIn=self.TIMEOUT)

    def publish(self, namespace, file_path, mime_type=None):
        file_path = ensure_path(file_path)
        file_name = safe_filename(file_path, default='data')
        store_path = '{0}/{1}'.format(namespace, file_name)
        extra = {}
        if mime_type is not None:
            extra['ContentType'] = mime_type
        with open(file_path, 'rb') as fh:
            self.client.upload_fileobj(fh, self.bucket, str(store_path),
                                       ExtraArgs=extra)

    def delete_publication(self, namespace, file_name):
        key = '{0}/{1}'.format(namespace, file_name)
        key = self._locate_key(prefix=key)
        if key is not None:
            self.client.delete_object(Bucket=self.bucket, Key=key)

    def generate_publication_url(self, namespace, file_name, mime_type=None,
                                 expire=None):
        key = '{0}/{1}'.format(namespace, file_name)
        key = self._locate_key(prefix=key)
        if key is None:
            return
        params = {
            'Bucket': self.bucket,
            'Key': key
        }
        if mime_type is not None:
            params['ResponseContentType'] = mime_type
        disposition = 'attachment; filename=%s' % file_name
        params['ResponseContentDisposition'] = disposition
        expire = expire or self.TIMEOUT
        return self.client.generate_presigned_url('get_object',
                                                  Params=params,
                                                  ExpiresIn=expire)



from unittest import TestCase

from servicelayer.cache import get_redis
from servicelayer.jobs import Job, Stage, Task
from servicelayer.reporting import Reporter, OP_REPORT, Status
from servicelayer.util import dump_json


class ReportingTest(TestCase):
    def setUp(self):
        self.conn = get_redis()
        self.dataset = 'my-dataset'
        self.operation = 'OP_FOO'
        self.job = Job.create(self.conn, self.dataset)

    def _get_task(self):
        """return the newly created reporting task"""
        return Stage.get_task(self.conn, [OP_REPORT])

    def test_reporting(self):
        payload = {
            'msg': 'I do report.'
        }

        reporter = Reporter(stage=self.job.get_stage(self.operation))

        # start
        reporter.start(**payload)
        task = self._get_task()

        self.assertEqual(task.stage.stage, OP_REPORT)

        data = task.payload
        values = {
            'msg': 'I do report.',
            'status': Status.START,
            'has_error': False,
            'operation': self.operation,
            'dataset': self.dataset,
            'updated_at': None,
            'start_at': None,
            'job': self.job.id
        }
        for key, value in values.items():
            self.assertIn(key, data)
            if value is not None:
                self.assertEqual(value, data[key])

        # success
        reporter.end(**payload)
        task = self._get_task()
        self.assertEqual(Status.SUCCESS, task.payload['status'])
        self.assertIn('success_at', task.payload)

        # failing
        reporter.error(Exception('An error occured!'), **payload)
        task = self._get_task()
        data = task.payload
        self.assertEqual(Status.ERROR, data['status'])
        self.assertTrue(data['has_error'])
        self.assertIn('error_at', data)
        self.assertEqual('Exception', data['error_name'])
        self.assertEqual('An error occured!', data['error_msg'])

    def test_reporting_from_task(self):
        task_data = {
            'dataset': self.dataset,
            'job': self.job.id,
            'stage': self.job.get_stage(self.operation).stage,
            'payload': {
                'entities': ['a', 'b', 'c']
            }
        }
        task = Task.unpack(self.conn, dump_json(task_data))
        serialized = task.serialize()
        reporter = Reporter(task=task)
        reporter.start()

        task = self._get_task()
        self.assertEqual(task.stage.stage, OP_REPORT)

        data = task.payload
        values = {
            'status': Status.START,
            'has_error': False,
            'operation': self.operation,
            'dataset': self.dataset,
            'updated_at': None,
            'start_at': None,
            'job': self.job.id,
            'task': serialized
        }
        for key, value in values.items():
            self.assertIn(key, data)
            if value is not None:
                self.assertEqual(value, data[key])
